package com.kameleoon.kameleoon_client_flutter

import android.app.Activity
import android.content.Context
import android.util.Log
import androidx.annotation.NonNull
import com.kameleoon.*
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import org.json.JSONArray
import org.json.JSONObject
import java.util.*
import kotlin.math.exp


/** KameleoonClientFlutterPlugin */
class KameleoonClientFlutterPlugin: FlutterPlugin, MethodCallHandler, ActivityAware {
  /// The MethodChannel that will the communication between Flutter and native Android
  ///
  /// This local reference serves to register the plugin with the Flutter Engine and unregister it
  /// when the Flutter Engine is detached from the Activity
  private enum class Method(override val value: String): RawRepresentable<String> {
    CREATE("create"), FORGET("forget"),
    RUN_WHEN_READY("runWhenReady"),
    GET_VISITOR_CODE("getVisitorCode"),
    TRIGGER_EXPERIMENT("triggerExperiment"), GET_VARIATION_ASSOCIATED_DATA("getVariationAssociatedData"),
    ADD_DATA("addData"), FLUSH("flush"), TRACK_CONVERSION("trackConversion"),
    GET_FEATURE_VARIATION_KEY("getFeatureVariationKey"), IS_FEATURE_ACTIVE("isFeatureActive"),
    GET_FEATURE_VARIABLE("getFeatureVariable"), GET_FEATURE_ALL_VARIABLES("getFeatureAllVariables"),
    GET_REMOTE_DATA("getRemoteData"),
    UPDATE_CONFIGURATION_HANDLER("updateConfigurationHandler"),
    GET_EXPERIMENT_LIST("getExperimentList"), GET_FEATURE_LIST("getFeatureList"),
    GET_FEATURE_LIST_ACTIVE("getFeatureListActive")
  }

  private enum class Argument(override val value: String): RawRepresentable<String> {
    SITE_CODE("siteCode"), CLIENT_ID("clientId"), CLIENT_SECRET("clientSecret"), ENVIRONMENT("environment"),
    REFRESH_INTERVAL_MINUTES("refreshIntervalMinutes"), VISITOR_CODE("visitorCode"),
    EXPERIMENT_ID("experimentId"), VARIATION_ID("variationId"), ONLY_ALLOCATED("onlyAllocated"),
    FEATURE_KEY("featureKey"), VARIATION_KEY("variationKey"), VARIABLE_KEY("variableKey"),
    DATA_LIST("dataList"), DATA_TYPE("dataType"), ID("id"), VALUES("values"), URL("url"), TITLE("title"),
    REFERRERS("referrers"), DEVICE("device"), BROWSER("browser"),
    TIMEOUT("timeout"),
    GOAL_ID("goalId"), REVENUE("revenue"), NEGATIVE("negative"),
    KEY("key"),
    SET("set")
  }

  private enum class ErrorCode(override val value: String): RawRepresentable<String> {
    INTERACTION("flutter_interaction_error"),
    SDK_NOT_READY("sdk_not_ready"),
    CREDENTIALS_NOT_FOUND("credentials_not_found"),
    SITE_CODE_REQUIRED("sitecode_required"), SITE_CODE_DISABLED("sitecode_disabled"),
    EXPERIMENT_NOT_FOUND("experiment_not_found"), FEATURE_NOT_FOUND("feature_not_found"),
    NOT_TARGETED("not_targeted"), NOT_ALLOCATED("not_allocated"),
    VARIATION_NOT_FOUND("variation_not_found"), VARIABLE_NOT_FOUND("variable_not_found"),
    VISITOR_CODE_NOT_VALID("visitor_code_not_valid"),
    UNDEFINED("undefined")
  }

  private enum class CustomDataType(override val value: String): RawRepresentable<String> {
    CUSTOM_DATA("CustomData"), PAGE_VIEW("PageView"), DEVICE("Device"), CONVERSION("Conversion"),
    BROWSER("Browser")
  }

  private val VISITOR_CODE_KEY = "kameleoon_sdk_visitor_code"

  private lateinit var channel : MethodChannel

  private lateinit var activity: Activity
  private lateinit var context: Context
  private val clients = mutableMapOf<String, KameleoonClient>()

  override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
    channel = MethodChannel(flutterPluginBinding.binaryMessenger, "com.kameleoon.flutter")
    channel.setMethodCallHandler(this)
    context = flutterPluginBinding.applicationContext
  }

  override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
    val siteCode = fetch<String>(call, Argument.SITE_CODE) ?: run {
      result.error(ErrorCode.SITE_CODE_REQUIRED.value, "siteCode is required", null)
      return
    }
    print(call.method)
    val method = valueOf<Method, String>(call.method) ?: run {
      result.error(ErrorCode.INTERACTION.value, "${call.method} method not found ", null)
      return
    }

    if (method == Method.CREATE) {
      create(siteCode, call, result)
    } else {
      val client = clients[siteCode] ?: run {
        catchException(KameleoonException.SDKNotReady("SDK is not ready"), result)
        return
      }
      when (method) {
        Method.FORGET -> forget(siteCode)
        Method.RUN_WHEN_READY -> runWhenReady(client, siteCode, call, result)
        Method.TRIGGER_EXPERIMENT -> triggerExperiment(client, call, result)
        Method.GET_VARIATION_ASSOCIATED_DATA -> getVariationAssociatedData(client, call, result)
        Method.FLUSH -> flush(client, call, result)
        Method.TRACK_CONVERSION -> trackConversion(client, call, result)
        Method.GET_FEATURE_VARIATION_KEY -> getFeatureVariationKey(client, call, result)
        Method.GET_FEATURE_VARIABLE -> getFeatureVariable(client, call, result)
        Method.IS_FEATURE_ACTIVE -> isFeatureActive(client, call, result)
        Method.GET_FEATURE_ALL_VARIABLES -> getFeatureAllVariables(client, call, result)
        Method.GET_VISITOR_CODE -> getVisitorCode(client, call, result)
        Method.ADD_DATA -> addData(client, call, result)
        Method.GET_EXPERIMENT_LIST -> getExperimentList(client, call, result)
        Method.GET_FEATURE_LIST -> getFeatureList(client, result)
        Method.GET_FEATURE_LIST_ACTIVE -> getFeatureListActive(client, call, result)
        Method.GET_REMOTE_DATA -> getRemoteData(client, call, result)
        Method.UPDATE_CONFIGURATION_HANDLER -> updateConfigurationHandler(client, siteCode, call, result)
        else ->
          result.error(ErrorCode.INTERACTION.value, "${call.method} method not found ", null)
      }
    }
  }

  private fun create(siteCode: String, call: MethodCall, @NonNull result: Result) {
    try {
      val clientId = fetch(call, Argument.CLIENT_ID) ?: ""
      val clientSecret = fetch(call, Argument.CLIENT_SECRET) ?: ""
      if (clientId.isBlank() || clientSecret.isBlank()) {
        throw KameleoonException.CredentialsNotFound("clientId or clientSecret are empty or invalid")
      }
      val environment = fetch<String>(call, Argument.ENVIRONMENT)
      val refreshIntervalMinutes = checkValue(call, Argument.REFRESH_INTERVAL_MINUTES) ?: 60
      val configBuilder = KameleoonConfiguration.Builder()
        .clientId(clientId)
        .clientSecret(clientSecret)
        .configurationRefreshInterval(refreshIntervalMinutes)
      environment?.takeIf { it.isNotBlank() }?.let {
        configBuilder.environment(it)
      }
      val client = KameleoonClientFactory.create(siteCode, configBuilder.build(), context)
      clients[siteCode] = client
      result.success("")
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun forget(siteCode: String) {
    KameleoonClientFactory.forget(siteCode)
    clients.remove(siteCode)
  }

  private fun runWhenReady(client: KameleoonClient, siteCode: String, call: MethodCall, result: Result) {
    val timeout = checkValue(call, Argument.TIMEOUT) ?: 2_000
    client.runWhenReady(object : KameleoonReadyCallback {
      override fun onReady() {
        activity.runOnUiThread {
          channel.invokeMethod("setReady", mapOf("siteCode" to siteCode, "ready" to true))
        }
      }

      override fun onTimeout() {
        activity.runOnUiThread {
          channel.invokeMethod("setReady", mapOf("siteCode" to siteCode, "ready" to false))
        }
      }
    }, timeout)
    result.success("")
  }

  private fun triggerExperiment(client: KameleoonClient, call: MethodCall, result: Result) {
    val visitorCode = checkValue(call, Argument.VISITOR_CODE) ?: ""
    val experimentId = checkValue(call, Argument.EXPERIMENT_ID) ?: 0
    try {
      val variationId = client.triggerExperiment(visitorCode, experimentId)
      result.success(variationId)
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun getVariationAssociatedData(client: KameleoonClient, call: MethodCall, result: Result) {
    val variationId = checkValue(call, Argument.VARIATION_ID) ?: 0
    try {
      val obj = client.getVariationAssociatedData(variationId)?.toMap() ?: mapOf<String, String>()
      result.success(obj)
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun flush(client: KameleoonClient, call: MethodCall, result: Result) {
    val visitorCode = checkValue(call, Argument.VISITOR_CODE) ?: ""
    try {
      if (visitorCode.isNotBlank()) {
        client.flush(visitorCode)
      } else {
        client.flush(null)
      }
      result.success("")
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun trackConversion(client: KameleoonClient, call: MethodCall, result: Result) {
    val visitorCode = checkValue(call, Argument.VISITOR_CODE) ?: ""
    val goalId = checkValue(call, Argument.GOAL_ID) ?: 0
    val revenue = checkValue(call, Argument.REVENUE) ?: 0.0
    try {
      client.trackConversion(visitorCode, goalId, revenue.toFloat())
      result.success("")
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun getFeatureVariationKey(client: KameleoonClient, call: MethodCall, result: Result) {
    val visitorCode = checkValue(call, Argument.VISITOR_CODE) ?: ""
    val featureKey = checkValue(call, Argument.FEATURE_KEY) ?: ""
    try {
      val variationKey = client.getFeatureVariationKey(visitorCode, featureKey)
      result.success(variationKey)
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun getFeatureVariable(client: KameleoonClient, call: MethodCall, result: Result) {
    val visitorCode = checkValue(call, Argument.VISITOR_CODE) ?: ""
    val featureKey = checkValue(call, Argument.FEATURE_KEY) ?: ""
    val variableKey = checkValue(call, Argument.VARIABLE_KEY) ?: ""
    try {
      val variable = client.getFeatureVariable(visitorCode, featureKey, variableKey)
      result.success(variable?.toFlutterResult())
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun isFeatureActive(client: KameleoonClient, call: MethodCall, result: Result) {
    val visitorCode = checkValue(call, Argument.VISITOR_CODE) ?: ""
    val featureKey = checkValue(call, Argument.FEATURE_KEY) ?: ""
    try {
      val active = client.isFeatureActive(visitorCode, featureKey)
      result.success(active)
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun getFeatureAllVariables(client: KameleoonClient, call: MethodCall, result: Result) {
    val featureKey = checkValue(call, Argument.FEATURE_KEY) ?: ""
    val variationKey = checkValue(call, Argument.VARIATION_KEY) ?: ""
    try {
      val allVariables = client.getFeatureAllVariables(featureKey, variationKey)
      result.success(allVariables.toFlutterResult())
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun getVisitorCode(client: KameleoonClient, call: MethodCall, result: Result) {
    val prefs = activity.getSharedPreferences(VISITOR_CODE_KEY, Context.MODE_PRIVATE)
    val visitorCode = prefs.getString(VISITOR_CODE_KEY, null) ?: UUID.randomUUID().toString().also {
      prefs.edit().putString(VISITOR_CODE_KEY, it).apply()
    }
    result.success(visitorCode)
  }

  private fun addData(client: KameleoonClient, call: MethodCall, result: Result) {
    val method = Method.ADD_DATA
    val visitorCode = checkValue(call, Argument.VISITOR_CODE) ?: ""
    val dataList = checkValue<List<Map<String,Any>>>(call, Argument.DATA_LIST) ?: listOf()
    val datas = dataList.mapNotNull {
      val dataType = checkValue(it, Argument.DATA_TYPE, method) ?: ""
      when(valueOf<CustomDataType, String>(dataType)) {
        CustomDataType.CUSTOM_DATA -> {
          val id = checkValue(it, Argument.ID, method) ?: 0
          val values = checkValue(it, Argument.VALUES, method) ?: listOf<String>()
          Data.CustomData(id, values)
        }
        CustomDataType.PAGE_VIEW -> {
          val url = checkValue(it, Argument.URL, method) ?: ""
          val title = checkValue(it, Argument.TITLE, method) ?: ""
          val referrers = checkValue(it, Argument.REFERRERS, method) ?: listOf<Int>()
          Data.PageView(url, title, referrers)
        }
        CustomDataType.DEVICE -> {
          val device = checkValue(it, Argument.DEVICE, method) ?: ""
          valueOf<Data.Device>(device) ?: Data.Device.PHONE
        }
        CustomDataType.BROWSER -> {
          val index = checkValue(it, Argument.BROWSER, method) ?: 0
          enumValues<Data.Browser>().firstOrNull { br -> br.ordinal == index } ?: Data.Browser.CHROME
        }
        CustomDataType.CONVERSION -> {
          val goalId = checkValue(it, Argument.GOAL_ID, method) ?: 0
          val revenue = checkValue(it, Argument.REVENUE, method) ?: 0.0
          val negative = checkValue(it, Argument.NEGATIVE, method) ?: false
          Data.Conversion(goalId, revenue.toFloat(), negative)
        }
        else -> {
          print("Data type not compatible: $dataType")
          null
        }
      }
    }
    try {
      datas.forEach {
        client.addData(visitorCode, it)
      }
      result.success("")
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun getExperimentList(client: KameleoonClient, call: MethodCall, result: Result) {
    val visitorCode = fetch(call, Argument.VISITOR_CODE) ?: ""
    val onlyAllocated = fetch(call, Argument.ONLY_ALLOCATED) ?: true
    if (visitorCode.isEmpty()) {
      result.success(client.experimentList)
    } else {
      try {
        val experimentList = client.getExperimentListForVisitorCode(visitorCode, onlyAllocated)
        result.success(experimentList)
      } catch (e: Exception) {
        catchException(e, result)
      }
    }
  }

  private fun getFeatureList(client: KameleoonClient, result: Result) {
    result.success(client.featureList)
  }

  private fun getFeatureListActive(client: KameleoonClient, call: MethodCall, result: Result) {
    val visitorCode = checkValue(call, Argument.VISITOR_CODE) ?: ""
    try {
      val featureList = client.getActiveFeatureListForVisitorCode(visitorCode)
      result.success(featureList)
    } catch (e: Exception) {
      catchException(e, result)
    }
  }

  private fun getRemoteData(client: KameleoonClient, call: MethodCall, result: Result) {
    val key = checkValue(call, Argument.KEY) ?: ""
    client.getRemoteData(key, object: KameleoonDataCallback {
      override fun onSuccess(jsonObject: JSONObject) {
        activity.runOnUiThread {
          result.success(jsonObject.toMap())
        }
      }

      override fun onFail(e: Exception) {
        activity.runOnUiThread {
          catchException(e, result)
        }
      }
    })
  }

  private fun updateConfigurationHandler(client: KameleoonClient, siteCode: String, call: MethodCall, result: Result) {
    val set = checkValue(call, Argument.SET) ?: false
    if (set) {
      client.updateConfigurationHandler {
        activity.runOnUiThread {
          channel.invokeMethod("updateConfigurationHandler", mapOf("siteCode" to siteCode))
        }
      }
    } else {
      client.updateConfigurationHandler(null)
    }
    result.success("")
  }

  override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
    channel.setMethodCallHandler(null)
  }

  override fun onAttachedToActivity(binding: ActivityPluginBinding) {
    activity = binding.activity
  }

  override fun onDetachedFromActivityForConfigChanges() {
    TODO("Not yet implemented")
  }

  override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
    TODO("Not yet implemented")
  }

  override fun onDetachedFromActivity() {
//    TODO("Not yet implemented")
  }

  private fun catchException(exception: Exception, result: Result) {
    val errorCode: ErrorCode = when (exception) {
      is KameleoonException.ExperimentConfigurationNotFound -> ErrorCode.EXPERIMENT_NOT_FOUND
      is KameleoonException.FeatureConfigurationNotFound -> ErrorCode.FEATURE_NOT_FOUND
      is KameleoonException.SDKNotReady -> ErrorCode.SDK_NOT_READY
      is KameleoonException.NotTargeted -> ErrorCode.NOT_TARGETED
      is KameleoonException.NotAllocated -> ErrorCode.NOT_ALLOCATED
      is KameleoonException.VariationConfigurationNotFound -> ErrorCode.VARIATION_NOT_FOUND
      is KameleoonException.FeatureVariableNotFound -> ErrorCode.VARIABLE_NOT_FOUND
      is KameleoonException.VisitorCodeNotValid -> ErrorCode.VISITOR_CODE_NOT_VALID
      is KameleoonException.SiteCodeDisabled -> ErrorCode.SITE_CODE_DISABLED
      is KameleoonException.CredentialsNotFound -> ErrorCode.CREDENTIALS_NOT_FOUND
      else -> ErrorCode.UNDEFINED
    }
    result.error(errorCode.value, exception.message, exception.stackTrace.toString())
  }

  private inline fun <reified T> fetch(call: MethodCall, key: Argument): T? {
    return call.argument<T>(key.value)
  }

  private inline fun <reified T> checkValue(call: MethodCall, key: Argument): T? {
    val value = fetch<T>(call, key)
    if (value == null) {
      val methodName = valueOf<Method, String>(call.method)?.value ?: ""
      println("Kameleoon SDK: Unexpected null for method: ${methodName}, value: ${key.value}")
    }
    return value
  }

  private inline fun <reified T> checkValue(dict: Map<String, Any>, key: Argument, method: Method): T? {
    val value = dict[key.value]
    if (value == null) {
      println("Kameleoon SDK: Unexpected null for method: ${method.value}, value: ${key.value}")
    }
    return value as? T
  }
}