import 'package:collection/collection.dart';

import 'package:flutter/foundation.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:integration_test/integration_test.dart';
import 'package:kameleoon_client_flutter/src/kameleoon_client.dart';
import 'package:kameleoon_client_flutter/src/kameleoon_client_factory.dart';
import 'package:kameleoon_client_flutter/src/kameleoon_data.dart';
import 'package:kameleoon_client_flutter/src/kameleoon_exception.dart';
import 'package:kameleoon_client_flutter/src/kameleoon_configuration.dart';

import 'constants.dart';
import 'utils.dart';

void main() {
  IntegrationTestWidgetsFlutterBinding.ensureInitialized();

  late final KameleoonClient client;
  late final KameleoonClient clientDisabled;

  setUpAll(() {
    final config = KameleoonConfiguration("clientId", "clientSecret");
    client = KameleoonClientFactory.create(SITECODE, config);
    clientDisabled = KameleoonClientFactory.create(SITECODE_DISABLED, config);
  });

  testWidgets("runWhenReadyTimeout", (WidgetTester tester) async {
    await testAPI(client);
    if (!kIsWeb) { // TODO: fix when will be available in JS SDK
      await testAPI(clientDisabled);
    }
  });

  testWidgets("isReady", (WidgetTester tester) async {
    await testAPI(client, () {
      expect(client.isReady(), equals(true), reason: makeReason("client.isReady() != true"));
    });
    await testAPI(clientDisabled, () {
      expect(clientDisabled.isReady(), equals(true), reason: makeReason("client.isReady() != true"));
    });
  });

  testWidgets("getFeatureVariable__all_types", (WidgetTester tester) async {
    await testAPI(client, () async {
      const visitorCode = "testGetFeatureVariable";
      const featureKey = "test_feature_variables";
      const boolTrue = "boolTrueKey", boolFalse = "boolFalseKey", number = "numberKey", numberNull = "numberNullKey",
        string = "stringKey", json = "jsonKey", jsonArray = "jsonArrayKey";

      final variableTrue = await client.getFeatureVariable(visitorCode, featureKey, boolTrue);
      final variableFalse = await client.getFeatureVariable(visitorCode, featureKey, boolFalse);
      final variableNumber = await client.getFeatureVariable(visitorCode, featureKey, number);
      final variableNumberNull = await client.getFeatureVariable(visitorCode, featureKey, numberNull);
      final variableString = await client.getFeatureVariable(visitorCode, featureKey, string);

      expect(variableTrue, equals(true), reason: makeReason("variableByIdTrue != true"));
      expect(variableFalse, equals(false), reason: makeReason("variableByIdFalse != false"));
      expect(variableNumber, equals(100), reason: makeReason("variableByIdInt != 100"));
      expect(variableNumberNull, equals(0), reason: makeReason("variableByIdIntNull != 0"));
      expect(variableString, equals("TestString"), reason: makeReason("variableByIdString != TestString"));

      final variableByIdJSON = await client.getFeatureVariable(visitorCode, featureKey, json) as Map;
      final variableByIdArrayJSON = await client.getFeatureVariable(visitorCode, featureKey, jsonArray) as List;
      expect(variableByIdJSON['test'], equals('json'), reason: makeReason("variableByIdJSON['test'] != 'json'"));
      expect(variableByIdArrayJSON.length > 1 ? (variableByIdArrayJSON[1] as Map)['key'] : "", equals('value-2'), reason: makeReason("(variableByIdArrayJSON[1] as Map<String, Object>)['key'] != 'value-2'"));

      const expectedVariableValue = "toto";
      const variableKey = "toto";
      const featureKeyTestFF = "testff";
      final variable139626ByKey = await client.getFeatureVariable(visitorCode, featureKeyTestFF, variableKey);

      expect(expectedVariableValue, equals(variable139626ByKey), reason: makeReason("expectedVariableValue != variable139626ByKey"));

      const fearureKeyNotFound = "not_found_feaure_key";
      const variableKeyNotFound = "not_found_variable_key";
      const reasonDesc = "getFeatureVariable must throw FeatureConfigurationNotFound";

      expect(
        client.getFeatureVariable(visitorCode, fearureKeyNotFound, variableKey),
        throwsA(isInstanceOf<FeatureNotFound>()),
        reason: makeReason(reasonDesc));
      expect(
        client.getFeatureVariable(visitorCode, featureKey, variableKeyNotFound),
        throwsA(isInstanceOf<VariableNotFound>()),
        reason: makeReason(reasonDesc));
    });
  });

  testWidgets("getFeatureAllVariables", (WidgetTester tester) async {
    if (kIsWeb) { // TODO: fix when will be available in JS SDK
      print("Method isn't supported by web, will be added later");
      return;
    }
    const featureKey = "test_feature_variables";
    const featureKeyErr = "featureKeyErr";
    const variationKey = "on";
    const variationKeyErr = "variationKeyErr";

    const boolTrue = "boolTrueKey", boolFalse = "boolFalseKey", number = "numberKey", numberNull = "numberNullKey",
      string = "stringKey", json = "jsonKey", jsonArray = "jsonArrayKey";

    expect(client.getFeatureAllVariables(featureKeyErr, variationKey), throwsA(isInstanceOf<FeatureNotFound>()));
    expect(client.getFeatureAllVariables(featureKey, variationKeyErr), throwsA(isInstanceOf<VariationNotFound>()));

    var allVariables = await client.getFeatureAllVariables(featureKey, variationKey);

    expect(allVariables[boolTrue], equals(true), reason: makeReason("allVariables[boolTrue]"));
    expect(allVariables[boolFalse], equals(false), reason: makeReason("allVariables[boolFalse]"));
    expect(allVariables[number], equals(100), reason: makeReason("allVariables[number]"));
    expect(allVariables[numberNull], equals(0), reason: makeReason("allVariables[numberNull]"));
    expect(allVariables[string], equals("TestString"), reason: makeReason("allVariables[string]"));
    expect(allVariables[json]['test'], equals("json"), reason: makeReason("allVariables[json]['test']"));
    expect(allVariables[jsonArray][1]['key'],
      equals("value-2"), reason: makeReason("allVariables[jsonArray][1]['key']"));
  });

  testWidgets("getVariationAssociatedData", (WidgetTester tester) async {
    await testAPI(client, () async {
      const mockedVariationId = 617899;

      final data = await client.getVariationAssociatedData(mockedVariationId);

      final toto = data["toto"] as bool;
      final titi = data["titi"] as String;
      final tutu = data["tutu"] as bool;

      expect(toto && titi == 'tata' && !tutu, equals(true), reason: makeReason("(toto && titi == 'tata' && !tutu) == false") );

      const variationIdNotFound = 617899000;
      const reasonDesc = "obtainVariationAssociatedData must throw VariationConfigurationNotFound";
      expect(
        client.getVariationAssociatedData(variationIdNotFound),
        throwsA(isInstanceOf<VariationNotFound>()),
        reason: makeReason(reasonDesc));
    });
  });

  testWidgets("isFeatureActiveNotActive", (WidgetTester tester) async {
    await testAPI(client, () async {
      const featureKey = "testff";
      expect(await client.isFeatureActive(VISITOR_CODE, featureKey),
        equals(false), reason: makeReason("activateFeature != null"));
    });
  });

  testWidgets("trackConversion", (WidgetTester tester) async {
    await testAPI(client, () async {
      const goalId = 238420;
      await client.trackConversion(VISITOR_CODE, goalId, 10);
      await client.trackConversion(VISITOR_CODE, goalId);
    });
  });

  testWidgets("triggerExperiments", (WidgetTester tester) async {
    await testAPI(client, () async {
      const visitorCode_2 = "test";
      const experimentId_1 = 136200, experimentId_2 = 136204, experimentId_3 = 136226;
      const expectedValue_1 = [0, 617899], expectedValue_2 = [617909, 617909], expectedValue_3 = [617982, 617983];
      const experimentId_Targeting = 136215;
      const expectedValue_Targeting = 0;
      final customData = CustomData(0, "true");

      final id136200_1 = await client.triggerExperiment(VISITOR_CODE, experimentId_1);
      final id136200_2 = await client.triggerExperiment(visitorCode_2, experimentId_1);
      expect(listEquals([id136200_1, id136200_2], expectedValue_1), equals(true), reason: makeReason("id136200_1, id136200_2] != expectedValue_1"));

      final id136204_1 = await client.triggerExperiment(VISITOR_CODE, experimentId_2);
      final id136204_2 = await client.triggerExperiment(visitorCode_2, experimentId_2);
      expect(listEquals([id136204_1, id136204_2], expectedValue_2), equals(true), reason: makeReason("id136204_1, id136204_2] != expectedValue_2"));

      final id136226_1 = await client.triggerExperiment(VISITOR_CODE, experimentId_3);
      final id136226_2 = await client.triggerExperiment(visitorCode_2, experimentId_3);
      expect(listEquals([id136226_1, id136226_2], expectedValue_3), equals(true), reason: makeReason("id136226_1, id136226_2] != expectedValue_3"));

      expect(client.triggerExperiment(visitorCode_2, experimentId_Targeting), throwsA(isInstanceOf<NotTargeted>()));
      await client.addData(visitorCode_2, [customData]);
      final id136215 = await client.triggerExperiment(visitorCode_2, experimentId_Targeting);
      expect(id136215, equals(expectedValue_Targeting), reason: makeReason("id136215 != expectedValue_Targeting"));
    });
  });

  testWidgets("trackConversion", (WidgetTester tester) async {
    await testAPI(client, () async {
      const goalId = 238420;
      await client.trackConversion(VISITOR_CODE, goalId, 10);
      await client.trackConversion(VISITOR_CODE, goalId);
    });
  });

  testWidgets("activateFeatureAddData", (WidgetTester tester) async {
      await testAPI(client, () async {
        const featureKey = "test_feature_variables";
        final customData = CustomData(2, "true");
        await client.addData(VISITOR_CODE, [customData]);
        expect(await client.isFeatureActive(VISITOR_CODE, featureKey),
          equals(true), reason: makeReason("activateFeature != null"));
      });
  });

  testWidgets("flush", (WidgetTester tester) async {
    await testAPI(client, () async {
      await client.addData(VISITOR_CODE, [
        CustomData(10, "afd"),
        CustomData(1, "https://google.com?key=\"aaaaa\"&value=10"),
        Browser(Browsers.CHROME),
        PageView("dfjksjdf", "title", [1231]),
        Device(Devices.PHONE)
      ]);
      await client.flush();
    });
  });

  testWidgets("visitorCode", (WidgetTester tester) async {
    await testAPI(client, () async {
      const VISITOR_CODE_MAX_LENGTH = 255;
      final visitorCodeTheLengthLimit = "a" * VISITOR_CODE_MAX_LENGTH;
      final visitorCodeTooLong = "a" * (VISITOR_CODE_MAX_LENGTH + 1);
      const visitorCodeEmpty = "";

      const mockedGoalId = 238420;
      final customData = CustomData(2, "true");
      const experimentId = 136200;
      const featureKey = "test_feature_variables";
      const variableKey = "boolTrueKey";

      checkVisitorCodeNotValid(String exception, Function testFunc) async {
        try {
          await testFunc.call();
          if(exception.isNotEmpty) {
            fail(makeReason("section '$exception' must throw expection"));
          }
        } on VisitorCodeNotValid catch(e) {
          expectSync(e.toString(), contains(exception), reason: makeReason("section '$exception'\n${e.toString()}" +
            "must contain '$exception'\nwrong throw VisitorCodeNotValid expection"));
        } catch (e) {
          fail(makeReason("section '$exception'\n${e.toString()}" +
            "must contain $exception\nthrow non VisitorCodeNotValid expection"));
        }
      }

      final mapLengthError = {
        visitorCodeTheLengthLimit: "",
        visitorCodeTooLong: "255",
        visitorCodeEmpty: "empty"
      };

      mapLengthError.forEach((visitorCode, exception) {
        checkVisitorCodeNotValid(exception, () async => await client.trackConversion(visitorCode, mockedGoalId) );
        checkVisitorCodeNotValid(exception, () async => await client.trackConversion(visitorCode, mockedGoalId, 10) );
        checkVisitorCodeNotValid(exception, () async => await client.addData(visitorCode, [customData]) );
        checkVisitorCodeNotValid(exception, () async => await client.isFeatureActive(visitorCode, featureKey) );
        checkVisitorCodeNotValid(exception, () async => await client.getFeatureVariationKey(visitorCode, featureKey) );
        checkVisitorCodeNotValid(exception,
          () async => await client.getFeatureVariable(visitorCode, featureKey, variableKey) );
        checkVisitorCodeNotValid(exception, () async => await client.triggerExperiment(visitorCode, experimentId) );
      });
    });
  });

  testWidgets("targeting", (WidgetTester tester) async {
      await testAPI(client, () async {
        const mockedFeatureKey = "test_first_level_or_operator";
        const mockedFeatureKeySecond = "copy_test_first_level_or_operator";
        const mockedFeatureKeyThird = "copy_copy_test_first_level_or_operator";

        expect(await client.isFeatureActive(VISITOR_CODE, mockedFeatureKey), equals(false));
        expect(await client.isFeatureActive(VISITOR_CODE, mockedFeatureKeySecond), equals(false));
        expect(await client.isFeatureActive(VISITOR_CODE, mockedFeatureKeyThird), equals(false));

        await client.addData(VISITOR_CODE, [CustomData(4, "true")]);
        expect(await client.isFeatureActive(VISITOR_CODE, mockedFeatureKey), equals(false));
        await client.addData(VISITOR_CODE, [CustomData(5, "false")]);
        expect(await client.isFeatureActive(VISITOR_CODE, mockedFeatureKey), equals(true));
        expect(await client.isFeatureActive(VISITOR_CODE, mockedFeatureKeySecond), equals(false));
        expect(await client.isFeatureActive(VISITOR_CODE, mockedFeatureKeyThird), equals(false));

        await client.addData(VISITOR_CODE, [CustomData(6, "999")]);
        expect(await client.isFeatureActive(VISITOR_CODE, mockedFeatureKeySecond), equals(true));
        expect(await client.isFeatureActive(VISITOR_CODE, mockedFeatureKeyThird), equals(false));

        await client.addData(VISITOR_CODE, [CustomData(7, "true")]);
        expect(await client.isFeatureActive(VISITOR_CODE, mockedFeatureKeyThird), equals(true));
      });
  });

  testWidgets("addingSameCustomData", (WidgetTester tester) async {
    await testAPI(client, () async {
      const featureFlagKey = "complextargeting";
      final customData900 = CustomData(1, "900");
      final customData999 = CustomData(1, "999");

      await client.addData(VISITOR_CODE, [customData900]);
      expect(await client.isFeatureActive(VISITOR_CODE, featureFlagKey), equals(false));
      await client.addData(VISITOR_CODE, [customData999]);
      expect(await client.isFeatureActive(VISITOR_CODE, featureFlagKey), equals(true));
    });
  });

  testWidgets("siteDisabled", (WidgetTester tester) async {
    if (!kIsWeb) { // TODO: fix when will be available in JS SDK
      await testAPI(clientDisabled, () async {
        const experimentIdNotFound = 0;
        const experimentId = 149355;

        expect(
          clientDisabled.triggerExperiment(VISITOR_CODE, experimentIdNotFound),
          throwsA(isInstanceOf<ExperimentNotFound>()),
          reason: makeReason("Error ExperimentNotFound"));
        expect(
          clientDisabled.triggerExperiment(VISITOR_CODE, experimentId),
          throwsA(isInstanceOf<SiteCodeDisabled>()),
          reason: makeReason("Error SiteCodeDisabled"));
      });
    }
});

  testWidgets("campaignNotFound", (WidgetTester tester) async {
    await testAPI(client, () async {
      const experimentIdNotFound = 0;
      const featureKeyNotFound = "featureKeyNotFound";

      expect(
        client.triggerExperiment(VISITOR_CODE, experimentIdNotFound),
        throwsA(isInstanceOf<ExperimentNotFound>()),
        reason: makeReason("Error ExperimentConfigurationNotFound"));
      expect(
        client.isFeatureActive(VISITOR_CODE, featureKeyNotFound),
        throwsA(isInstanceOf<FeatureNotFound>()),
        reason: makeReason("Error FeatureConfigurationNotFound"));
    });
  });

  testWidgets("getRemoteData", (WidgetTester tester) async {
    await testAPI(client, () async {
      const expectedValueTest1 = "some value";
      const expectedValueTest2 = "3";
      const expectedValueTest3 = "no value";

      var value = await client.getRemoteData("test-1");
      expect(value["some field to insert or update"],
        equals(expectedValueTest1), reason: makeReason("Error expectedValueTest1"));

      value = await client.getRemoteData("test-2");
      expect(value["hobbies"][2], equals(expectedValueTest2), reason: makeReason("Error expectedValueTest2"));

      value = await client.getRemoteData("test 3");
      expect(value["test-3"], equals(expectedValueTest3), reason: makeReason("Error expectedValueTest3"));
    });
  });

group("Experiment List", () {
  testWidgets("getExperimentList", (WidgetTester tester) async {
    await testAPI(client, () async {
      const expected = [136200, 136204, 136215, 136223, 136226, 146968, 157303, 161914, 162109, 162338];

      final actual = (await client.getExperimentList())..sort();

      expect(actual, equals(expected));
    });
  });

  testWidgets("getExperimentList_OnlyAllocated_False", (WidgetTester tester) async {
    await testAPI(client, () async {
      const visitorCode = "testGetExperimentListVisitor_OnlyAllocatedFalse";

      const expected = [136200, 136204, 136223, 136226, 146968, 161914];

      final actual = (await client.getExperimentList(visitorCode, false))..sort();

      expect(actual, equals(expected));
    });
  });

  testWidgets("getExperimentList_OnlyAllocated_True", (WidgetTester tester) async {
    await testAPI(client, () async {
      const visitorCodes = ["testGetExperimentListVisitor_OnlyAllocatedTrue",
        "testGetExperimentListVisitor_OnlyAllocatedTrue_2"];
      const expected = [[136200, 136204, 136223, 136226, 161914], [136200, 136204, 136226, 161914]];

      final actual = await Future.wait(visitorCodes.map((visitorCode) => client.getExperimentList(visitorCode)));
      actual[0].sort();
      actual[1].sort();

      expect(actual, equals(expected));
    });
  });
});

  group("Feature List", () {
    testWidgets("getFeatureList", (WidgetTester tester) async {
      await testAPI(client, () async {
        const expected = [
          "complextargeting", "complextargeting_2", "copy_copy_copy_test_first_level_or_operator",
          "copy_copy_test_first_level_or_operator", "copy_test_first_level_or_operator",
          "feature_flag_can_change_segment__no_tests_for_this_ff_","feature_flag_liga_stavok",
          "feature_flag_multi_environment_test", "feature_flag_test_real_time_configuration_service",
          "ff_new_rules", "test_feature_variables","test_first_level_or_operator", "test_pagination_1",
          "test_pagination_10", "test_pagination_11", "test_pagination_12", "test_pagination_2",
          "test_pagination_3_", "test_pagination_4", "test_pagination_5", "test_pagination_6",
          "test_pagination_7", "test_pagination_8", "test_pagination_9",
          "test_segment_lower_greater", "testff"
        ];

        final actual = await client.getFeatureList();

        expect(actual..sort(), equals(expected));
      });
    });

    testWidgets("getActiveFeatureListVisitor", (WidgetTester tester) async {
      await testAPI(client, () async {
        final visitorCode = VISITOR_CODE + "featurelistforvisitor";

        const expected = ["complextargeting_2", "feature_flag_multi_environment_test",
          "feature_flag_test_real_time_configuration_service"];

        final actual = await client.getFeatureListActive(visitorCode);

        expect(actual, equals(expected));
      });
    });
  });

  testWidgets("getFeatureVariationKey", (WidgetTester tester) async {
    await testAPI(client, () async {
      const featureKey = "ff_new_rules";
      final visitorCodes = ["b", "a", "d", "e"].map((e) => VISITOR_CODE + e);

      const expected = ["variation_1", "variation_2", "on", "off"];

      final actual = await Future.wait(visitorCodes.map((element) async =>
        await client.getFeatureVariationKey(element, featureKey)));

      expect(actual, equals(expected));
    });
  });

  testWidgets("getFeatureVariable", (WidgetTester tester) async {
    await testAPI(client, () async {
      const featureKey = "ff_new_rules";
      final visitorCodes = ["d", "b", "c", "e"].map((e) => VISITOR_CODE + e).toList();
      const variable = ["variable_number", "variable_string", "variable_bool", "variable_json"];

      const expected = [11.0, "variation", true, {"json": "off"}];

      final actual = await Future.wait(visitorCodes.mapIndexed((index, element) =>
        client.getFeatureVariable(element, featureKey, variable[index])));

      expect(actual, equals(expected));
    });
  });

}
