import Flutter
import UIKit
import kameleoonClient

public class SwiftKameleoonClientFlutterPlugin: NSObject, FlutterPlugin {
    private var clients = [String : KameleoonClient]()
    private var channel : FlutterMethodChannel

    private enum Method: String {
        case create, forget
        case runWhenReady
        case getVisitorCode
        case triggerExperiment, getVariationAssociatedData
        case addData, flush, trackConversion
        case getFeatureVariationKey, isFeatureActive, getFeatureVariable, getFeatureAllVariables
        case getRemoteData
        case updateConfigurationHandler
        case getExperimentList, getFeatureList, getFeatureListActive
    }
    
    private enum Argument: String {
        case siteCode, clientId, clientSecret, environment, refreshIntervalMinutes, visitorCode
        case experimentId, variationId, onlyAllocated
        case featureKey, variationKey, variableKey
        case dataList, dataType, id, values, url, title, referrers, device, browser
        case timeout
        case goalId, revenue, negative
        case key
        case set
    }
    
    private enum CustomDataType: String {
        case customData = "CustomData", pageView = "PageView", device = "Device", conversion = "Conversion",
             browser = "Browser"
    }
    
    private enum ErrorCode: String {
        case interaction = "flutter_interaction_error"
        case sdkNotReady = "sdk_not_ready"
        case credentialsNotFound = "credentials_not_found"
        case siteCodeRequired = "sitecode_required", siteCodeDisabled = "sitecode_disabled"
        case experimentNotFound = "experiment_not_found", featureNotFound = "feature_not_found"
        case notTargeted = "not_targeted", notAllocated = "not_allocated"
        case variationNotFound = "variation_not_found", variableNotFound = "variable_not_found"
        case visitorCodeNotValid = "visitor_code_not_valid"
        case undefined = "undefined"
    }

    
    private init(channel: FlutterMethodChannel) {
        self.channel = channel
    }

    public static func register(with registrar: FlutterPluginRegistrar) {
        let channel = FlutterMethodChannel(name: "com.kameleoon.flutter", binaryMessenger: registrar.messenger())
        let instance = SwiftKameleoonClientFlutterPlugin(channel: channel)
        registrar.addMethodCallDelegate(instance, channel: channel)
    }

    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        let arguments = call.arguments as? NSDictionary ?? [:]
        
        guard let siteCode = fetch(arguments, .siteCode) as String? else {
            let errorSiteCode = FlutterError.init(code: ErrorCode.siteCodeRequired.rawValue,
                                                  message: "siteCode is required",
                                                  details: nil)
            result(errorSiteCode)
            return
        }
        
        guard let method = Method(rawValue: call.method) else {
            let errorInteraction = FlutterError.init(code: ErrorCode.interaction.rawValue,
                                                     message: "method not found",
                                                     details: nil)
            result(errorInteraction)
            return
        }
        
        if (method == .create) {
            create(siteCode: siteCode, arguments: arguments, result: result)
        } else {
            guard let client = clients[siteCode] else {
                result(catchError(KameleoonError.sdkNotReady))
                return
            }
            switch (method) {
                case .forget:
                    forget(siteCode: siteCode)
                case .runWhenReady:
                    runWhenReady(client: client, siteCode: siteCode, arguments: arguments, result: result)
                case .triggerExperiment:
                    triggerExperiment(client: client, arguments: arguments, result: result)
                case .getVariationAssociatedData:
                    getVariationAssociatedData(client: client, arguments: arguments, result: result)
                case .flush:
                    flush(client: client, arguments: arguments, result: result)
                case .trackConversion:
                    trackConversion(client: client, arguments: arguments, result: result)
                case .getFeatureVariationKey:
                    getFeatureVariationKey(client: client, arguments: arguments, result: result)
                case .getFeatureVariable:
                    getFeatureVariable(client: client, arguments: arguments, result: result)
                case .isFeatureActive:
                    isFeatureActive(client: client, arguments: arguments, result: result)
                case .getFeatureAllVariables:
                    getFeatureAllVariables(client: client, arguments: arguments, result: result)
                case .getVisitorCode:
                    getVisitorCode(client: client, arguments: arguments, result: result)
                case .addData:
                    addData(client: client, arguments: arguments, result: result)
                case .getExperimentList:
                    getExperimentList(client: client, arguments: arguments, result: result)
                case .getFeatureList:
                    getFeatureList(client: client, result: result)
                case .getFeatureListActive:
                    getFeatureListActive(client: client, arguments: arguments, result: result)
                case .getRemoteData:
                    getRemoteData(client: client, arguments: arguments, result: result)
                case .updateConfigurationHandler:
                    updateConfigurationHandler(client: client, siteCode: siteCode, arguments: arguments, result: result)
                case .create:
                    break
            }
        }
    }
    
    private func forget(siteCode: String) {
        KameleoonClientFactory.forget(sitecode: siteCode)
        clients.removeValue(forKey: siteCode)
    }
    
    private func create(siteCode: String, arguments: NSDictionary, result: @escaping FlutterResult) {
        do {
            guard let clientId = fetch(arguments, .clientId) as String?,
                  let clientSecret = fetch(arguments, .clientSecret) as String? else {
                throw KameleoonError.credentialsNotFound
            }
            let environment = fetch(arguments, .environment) as String?
            let refreshIntervalMinutes = checkValue(arguments, .refreshIntervalMinutes, .create) as Int? ?? 60

            let config = try KameleoonClientConfig(clientId: clientId,
                                                   clientSecret: clientSecret,
                                                   refreshIntervalMinutes: refreshIntervalMinutes,
                                                   environment: environment)
            let kameleoonClient = try KameleoonClientFactory.create(sitecode: siteCode, config: config)
            clients[siteCode] = kameleoonClient
            result("")
        } catch {
            result(catchError(error))
        }
    }
    
    private func triggerExperiment(client: KameleoonClient, arguments: NSDictionary, result: @escaping FlutterResult) {
        let visitorCode = checkValue(arguments, .visitorCode, .triggerExperiment) as String? ?? ""
        let experimentId = checkValue(arguments, .experimentId, .triggerExperiment) as Int? ?? 0
        do {
            try result(client.triggerExperiment(visitorCode: visitorCode, experimentId: experimentId))
        } catch {
            result(catchError(error))
        }
    }
    
    private func flush(client: KameleoonClient, arguments: NSDictionary, result: @escaping FlutterResult) {
        do {
            let visitorCode = checkValue(arguments, .visitorCode, .triggerExperiment) as String? ?? ""
            if (!visitorCode.isEmpty){
                try client.flush(visitorCode: visitorCode)
            } else {
                try client.flush()
            }
            result("")
        } catch {
            result(catchError(error))
        }
    }
    
    private func trackConversion(client: KameleoonClient, arguments: NSDictionary, result: @escaping FlutterResult) {
        let visitorCode = checkValue(arguments, .visitorCode, .trackConversion) as String? ?? ""
        let goalId = checkValue(arguments, .goalId, .trackConversion) as Int? ?? 0
        let revenue = checkValue(arguments, .revenue, .trackConversion) as Double? ?? 0
        do {
            try client.trackConversion(visitorCode: visitorCode, goalId: goalId, revenue: Float(revenue))
            result("")
        } catch {
            result(catchError(error))
        }
    }
    
    private func getFeatureVariationKey(client: KameleoonClient,
                                        arguments: NSDictionary,
                                        result: @escaping FlutterResult) {
        let visitorCode = checkValue(arguments, .visitorCode, .getFeatureVariationKey) as String? ?? ""
        let featureKey = checkValue(arguments, .featureKey, .getFeatureVariationKey) as String? ?? ""
        do {
            let variationKey = try client.getFeatureVariationKey(visitorCode: visitorCode, featureKey: featureKey)
            result(variationKey)
        } catch {
            result(catchError(error))
        }
    }
    
    private func getFeatureVariable(client: KameleoonClient,
                                    arguments: NSDictionary,
                                    result: @escaping FlutterResult) {
        let visitorCode = checkValue(arguments, .visitorCode, .getFeatureVariable) as String? ?? ""
        let featureKey = checkValue(arguments, .featureKey, .getFeatureVariable) as String? ?? ""
        let variableKey = checkValue(arguments, .variableKey, .getFeatureVariable) as String? ?? ""
        do {
            let variable = try client.getFeatureVariable(visitorCode: visitorCode,
                                                       featureKey: featureKey,
                                                       variableKey: variableKey)
            result(variable)
        } catch {
            result(catchError(error))
        }
    }
    
    private func isFeatureActive(client: KameleoonClient, arguments: NSDictionary, result: @escaping FlutterResult) {
        let visitorCode = checkValue(arguments, .visitorCode, .isFeatureActive) as String? ?? ""
        let featureKey = checkValue(arguments, .featureKey, .isFeatureActive) as String? ?? ""
        do {
            let active = try client.isFeatureActive(visitorCode: visitorCode, featureKey: featureKey)
            result(active)
        } catch {
            result(catchError(error))
        }
    }
    
    private func getVisitorCode(client: KameleoonClient, arguments: NSDictionary, result: @escaping FlutterResult) {
        let VISITOR_CODE_KEY = "kameleoon_sdk_visitor_code"
        var visitorCode = UserDefaults.standard.string(forKey: VISITOR_CODE_KEY)
        if (visitorCode == nil) {
            visitorCode = UUID().uuidString
            UserDefaults.standard.setValue(visitorCode, forKey: VISITOR_CODE_KEY)
        }
        result(visitorCode)
    }
        
    private func getVariationAssociatedData(client: KameleoonClient,
                                            arguments: NSDictionary,
                                            result: @escaping FlutterResult) {
        let variationId = checkValue(arguments, .variationId, .getVariationAssociatedData) as Int? ?? 0
        do {
            let data = try client.getVariationAssociatedData(variationId: variationId)
            result(data)
        } catch {
            result(catchError(error))
        }
    }
    
    private func runWhenReady(client: KameleoonClient,
                              siteCode: String,
                              arguments: NSDictionary,
                              result: @escaping FlutterResult) {
        let timeout = checkValue(arguments, .timeout, .runWhenReady) as Int? ?? 2_000
        client.runWhenReady(timeoutMilliseconds: timeout) { success in
           if (success) {
               self.channel.invokeMethod("setReady", arguments: ["siteCode": siteCode, "ready": true])
           }
           else {
               self.channel.invokeMethod("setReady", arguments: ["siteCode": siteCode, "ready": false])
           }
        }
        result("")
    }
    
    private func addData(client: KameleoonClient, arguments: NSDictionary, result: @escaping FlutterResult) {
        let method = Method.addData
        let visitorCode = checkValue(arguments, .visitorCode, method) as String? ?? ""
        let dataList = checkValue(arguments, .dataList, method) as [NSDictionary]? ?? [[:]]
        let datas = dataList.compactMap { dataMap in
            let dataType = checkValue(dataMap, .dataType, method) as String? ??  ""
            let data: KameleoonData?
            switch CustomDataType(rawValue: dataType) {
                case .customData:
                    let id = checkValue(dataMap, .id, method) as Int? ?? 0
                    let values = checkValue(dataMap, .values, method) as [String]? ?? []
                    data = CustomData(id: id, values: values)
                case .pageView:
                    let url = checkValue(dataMap, .url, method) as String? ?? ""
                    let title = checkValue(dataMap, .title, method) as String? ?? ""
                    let referrers = checkValue(dataMap, .referrers, method) as [Int]? ?? []
                    data = PageView(url: url, title: title, referrers: referrers)
                case .device:
                    let device = checkValue(dataMap, .device, method) as String? ?? "PHONE"
                    data = Device(rawValue: device.lowercased())
                case .conversion:
                    let goalId = checkValue(dataMap, .goalId, method) as Int? ?? 0
                    let revenue = checkValue(dataMap, .revenue, method) as Float? ?? 0
                    let negative = checkValue(dataMap, .negative, method) as Bool? ?? false
                    data = Conversion(goalId: goalId, revenue: revenue, negative: negative)
                case .browser:
                    let browser = checkValue(dataMap, .browser, method) as Int? ?? 0
                    data = Browser.init(rawValue: browser)
                default:
                    data = nil
                    print("Data type not compatible: \(dataType)")
            }
            return data
        }
        do {
            try datas.forEach {
                try client.addData(visitorCode: visitorCode, data: $0);
            }
            result(true)
        } catch {
            result(catchError(error))
        }
    }
    
    private func getRemoteData(client: KameleoonClient, arguments: NSDictionary, result: @escaping FlutterResult) {
        let key = checkValue(arguments, .key, .getRemoteData) as String? ?? ""
        do {
            try client.getRemoteData(key: key) { [weak self] (data: Data) in
                do {
                    if let json = try JSONSerialization.jsonObject(with: data) as? [String: Any] {
                        result(json)
                    }
                } catch {
                    result(self?.catchError(error))
                }
            }
        } catch {
            result(catchError(error))
        }
    }
    
    private func getFeatureAllVariables(client: KameleoonClient,
                                        arguments: NSDictionary,
                                        result: @escaping FlutterResult) {
        let featureKey = checkValue(arguments, .featureKey, .getFeatureAllVariables) as String? ?? ""
        let variationKey = checkValue(arguments, .variationKey, .getFeatureAllVariables) as String? ?? ""
        do {
            let allVariables = try client.getFeatureAllVariables(featureKey: featureKey, variationKey: variationKey)
            result(allVariables)
        } catch {
            result(catchError(error))
        }
    }
    
    private func getExperimentList(client: KameleoonClient,
                                   arguments: NSDictionary,
                                   result: @escaping FlutterResult) {
        let visitorCode = fetch(arguments, .visitorCode) as String? ?? ""
        let onlyAllocated = fetch(arguments, .onlyAllocated) as Bool? ?? true
        if visitorCode.isEmpty {
            result(client.getExperimentList())
        } else {
            do {
                let experimentList = try client.getExperimentList(visitorCode: visitorCode,
                                                                  onlyAllocated: onlyAllocated)
                result(experimentList)
            } catch {
                result(catchError(error))
            }
        }
    }
    
    private func getFeatureList(client: KameleoonClient, result: @escaping FlutterResult) {
        result(client.getFeatureList())
    }
    
    private func getFeatureListActive(client: KameleoonClient,
                                      arguments: NSDictionary,
                                      result: @escaping FlutterResult) {
        let visitorCode = checkValue(arguments, .visitorCode, .getFeatureListActive) as String? ?? ""
        do {
            let featureList = try client.getActiveFeatureList(visitorCode: visitorCode)
            result(featureList)
        } catch {
            result(catchError(error))
        }
    }
    
    private func updateConfigurationHandler(client: KameleoonClient,
                                            siteCode: String,
                                            arguments: NSDictionary,
                                            result: @escaping FlutterResult) {
        let set = checkValue(arguments, .set, .updateConfigurationHandler) as Bool? ?? false
        if set {
            client.updateConfigurationHandler {
                self.channel.invokeMethod("updateConfigurationHandler", arguments: ["siteCode": siteCode])
            }
        } else {
            client.updateConfigurationHandler(nil)
        }
        result("")
    }
}

// MARK: - Fetch & Check Arguments

extension SwiftKameleoonClientFlutterPlugin {
    
    private func fetch<T>(_ arguments: NSDictionary, _ key: Argument) -> T? {
        return arguments[key.rawValue] as? T
    }
    
    private func checkValue<T>(_ arguments: NSDictionary, _ key: Argument, _ method: Method) -> T? {
        let value = fetch(arguments, key) as T?
        if value == nil {
            print("Kameleoon SDK: Unexpected nil for method: \(method.rawValue), value: \(key.rawValue)")
        }
        return value
    }
}

// MARK: - Error Handler

extension SwiftKameleoonClientFlutterPlugin {
    
    private func catchError(_ error: Error) -> FlutterError {
        let errorCode: ErrorCode
        switch error {
            case KameleoonError.experimentNotFound(_):
                errorCode = .experimentNotFound
            case KameleoonError.featureFlagNotFound(_):
                errorCode = .featureNotFound
            case KameleoonError.sdkNotReady:
                errorCode = .sdkNotReady
            case KameleoonError.notTargeted:
                errorCode = .notTargeted
            case KameleoonError.notAllocated(_, _):
                errorCode = .notAllocated
            case KameleoonError.variationNotFound(_):
                errorCode = .variationNotFound
            case KameleoonError.variableNotFound(_):
                errorCode = .variableNotFound
            case KameleoonError.visitorCodeNotValid:
                errorCode = .visitorCodeNotValid
            case KameleoonError.siteCodeDisabled:
                errorCode = .siteCodeDisabled
            case KameleoonError.credentialsNotFound:
                errorCode = .credentialsNotFound;
            default:
                errorCode = .undefined
        }
        return FlutterError.init(
            code: errorCode.rawValue,
            message: error.localizedDescription,
            details: nil)
    }
}
