import 'package:flutter/services.dart';
import '../kameleoon_client.dart';

import '../kameleoon_data.dart';
import '../kameleoon_exception.dart';
import 'kameleoon_client_callback.dart';
import '../kameleoon_configuration.dart';

/// Implementation [KameleoonClient] for iOS / Android platform
class KameleoonClientImpl extends KameleoonClient {
  MethodChannel _channel;

  bool ready = false;
  List<KameleoonClientCallback> _callbacks = [];
  String siteCode;
  Function? _handlerUpdateConfiguration;

  KameleoonClientImpl(this.siteCode, KameleoonConfiguration config, this._channel) {
    _channel.invokeMethod("create", <String, dynamic> {
      "siteCode": siteCode,
      "clientId": config.clientId,
      "clientSecret": config.clientSecret,
      "refreshIntervalMinutes": config.refreshIntervalMinutes,
      "environment": config.environment
    }).catchError(_handleException);
  }

  void forget() {
    _channel.invokeMethod("forget", <String, String> {
      "siteCode": siteCode
    });
  }

  ////////////////////////////////
  /// PUBLIC METHODS
  ////////////////////////////////

  Future<void> trackConversion(String visitorCode, int goalID, [double? revenue]) {
    return _channel.invokeMethod("trackConversion", <String, dynamic>{
      "visitorCode": visitorCode,
      "goalID": goalID,
      "revenue": revenue,
      "siteCode": this.siteCode}
    )
    .catchError(_handleException);
  }

  Future<int> triggerExperiment(String visitorCode, int experimentId) {
    return _channel.invokeMethod("triggerExperiment", <String, dynamic>{
        "visitorCode": visitorCode,
        "experimentId": experimentId,
        "siteCode": this.siteCode
      })
      .catchError(_handleException)
      .then((value) => value);
  }

  Future<dynamic> getVariationAssociatedData(int variationId) async {
    return _channel.invokeMethod(
      "getVariationAssociatedData", <String, dynamic>{
      "variationId": variationId,
      "siteCode": siteCode
    })
      .catchError(_handleException)
      .then((value) => value);
  }

  Future<dynamic> obtainVariationAssociatedData(int variationId) async => getVariationAssociatedData(variationId);

  Future<dynamic> obtainFeatureVariable(var featureKeyOrID, String variableKey) async {
    // "off" == default variation which always exists
    return getFeatureAllVariables(featureKeyOrID.toString(), "off").then((value) {
      if (value.containsKey(variableKey)) {
        return value[variableKey];
      } else {
        throw VariableNotFound(variableKey: variableKey);
      }
    });
  }

  bool? isReady() {
    return this.ready;
  }

  Future<void> addData(String visitorCode, List<Data> listData) async {
    var parameters = <String, Object>{};
    parameters["visitorCode"] = visitorCode;
    parameters["siteCode"] = siteCode;
    var dataList = List.generate(listData.length, (index) {
      Map<String, dynamic> dataToMap = {};
      var data = listData[index];
      dataToMap["dataType"] = data.dataType;
      switch (data.dataType) {
        case "CustomData":
          CustomData d = data as CustomData;
          dataToMap["id"] = d.id;
          dataToMap["values"] = d.values;
          break;
        case "PageView":
          PageView d = data as PageView;
          dataToMap["url"] = d.url;
          dataToMap["title"] = d.title;
          dataToMap["referrers"] = d.referrers;
          break;
        case "Conversion":
          Conversion d = data as Conversion;
          dataToMap["goalId"] = d.goalId;
          dataToMap["revenue"] = d.revenue;
          dataToMap["negative"] = d.negative;
          break;
        case "Browser":
          Browser b = data as Browser;
          dataToMap["browser"] = b.browser;
          break;
        case "Device":
          Device d = data as Device;
          dataToMap["device"] = d.device;
          break;
      }
      return dataToMap;
    }, growable: false);
    parameters["dataList"] = dataList;
    return _channel.invokeMethod("addData", parameters)
      .catchError(_handleException)
      .then((value) => value);
  }

  void runWhenReady(Function readyCallback, Function failCallback, [Duration? timeout]) async {
    if (this.ready) {
      readyCallback();
    } else {
      final runWhenReadyTimeout = timeout ?? Duration(seconds: 2);
      final callback = KameleoonClientCallback(readyCallback, failCallback);
      _callbacks.add(callback);
      _channel.invokeMethod("runWhenReady", <String, dynamic>{
          "timeout": runWhenReadyTimeout.inMilliseconds,
          "siteCode": this.siteCode})
        .catchError(_handleException)
        .timeout(runWhenReadyTimeout);
      Future.delayed(runWhenReadyTimeout, () => _callbackAfterStatusChange(false, callback));
    }
  }

  Future<void> flush([String? visitorCode]) async {
    var parameters = <String, dynamic>{
      "siteCode": siteCode,
    };
    if (visitorCode != null) {
      parameters["visitorCode"] = visitorCode;
    }
    return _channel.invokeMethod("flush", parameters)
      .catchError(_handleException);
  }

  Future<dynamic> getRemoteData(String key) async {
    var parameters = <String, dynamic>{
      "siteCode": siteCode,
      "key": key,
    };
    return _channel.invokeMethod("getRemoteData", parameters)
      .catchError(_handleException)
      .then((value) => value);
  }

  Future<dynamic> retrieveDataFromRemoteSource(String key) async => getRemoteData(key);

  Future<String> getFeatureVariationKey(String visitorCode, String featureKey) {
    return _channel.invokeMethod("getFeatureVariationKey", <String, String> {
      "siteCode": siteCode,
      "visitorCode": visitorCode,
      "featureKey": featureKey}
    )
    .catchError(_handleException)
    .then((value) => value);
  }

  Future<dynamic> getFeatureVariable(String visitorCode, String featureKey, String variableKey) {
    return _channel.invokeMethod("getFeatureVariable", <String, String> {
      "siteCode": siteCode,
      "visitorCode": visitorCode,
      "featureKey": featureKey,
      "variableKey": variableKey})
      .catchError(_handleException)
      .then((value) => value);

  }

  Future<bool> isFeatureActive(String visitorCode, var featureKeyOrID) async {
    return _channel.invokeMethod("isFeatureActive", <String, String> {
        "siteCode": siteCode,
        "visitorCode": visitorCode,
        "featureKey": featureKeyOrID}
    )
    .catchError(_handleException)
    .then((value) => value);
  }

  Future<Map<String, dynamic>> getFeatureAllVariables(String featureKey, String variationKey) {
    return _channel.invokeMethod("getFeatureAllVariables", <String, String> {
      "siteCode": siteCode,
      "featureKey": featureKey,
      "variationKey": variationKey}
    )
      .catchError(_handleException)
      .then((value) => Map.castFrom<dynamic, dynamic, String, Object>(value));
  }

  @override
  void updateConfigurationHandler(Function? handler) {
    _handlerUpdateConfiguration = handler;
    _channel.invokeMethod("updateConfigurationHandler", <String, Object> {
      "siteCode": siteCode,
      "set": _handlerUpdateConfiguration != null
    });
  }

  @override
  Future<List<int>> getExperimentList([String visitorCode = "", bool onlyAllocated = true]) {
    return _channel.invokeMethod("getExperimentList", <String, Object> {
      "siteCode": siteCode,
      "visitorCode": visitorCode,
      "onlyAllocated": onlyAllocated
    })
      .catchError(_handleException)
      .then((value) => List.castFrom<dynamic, int>(value));
  }

  @override
  Future<List<String>> getFeatureList() {
    return _channel.invokeMethod("getFeatureList", <String, String> {
      "siteCode": siteCode,
    })
      .then((value) => List.castFrom<dynamic, String>(value));
  }

  @override
  Future<List<String>> getFeatureListActive(String visitorCode) {
    return _channel.invokeMethod("getFeatureListActive", <String, String> {
      "siteCode": siteCode,
      "visitorCode": visitorCode
    })
      .catchError(_handleException)
      .then((value) => List.castFrom<dynamic, String>(value));
  }

  Future<String> getVisitorCode({String? domain, String? defaultVisitorCode}) {
    return _channel.invokeMethod("getVisitorCode", <String, String> {
      "siteCode": siteCode
    })
    .then((value) => value ?? defaultVisitorCode);
  }

  Future<String?> obtainVisitorCode([String? defaultVisitorCode]) async =>
    getVisitorCode(defaultVisitorCode: defaultVisitorCode);

  ////////////////////////////////
  /// PRIVATE METHODS
  ////////////////////////////////

  void _callbackAfterStatusChange(bool isReady, [KameleoonClientCallback? callback]) {
    List<int> idsToDelete = [];
    for (int i = 0; i < _callbacks.length; i++) {
      if (isReady) {
        _callbacks[i].readyCallback();
        idsToDelete.add(i);
      } else {
        if (_callbacks[i] == callback) {
          callback?.failCallback?.call();
          idsToDelete.add(i);
        }
      }
    }

    for (int i = idsToDelete.length - 1; i >= 0; i--) {
      _callbacks.removeAt(i);
    }
  }

  void nativeMethodCallHandler(MethodCall call) {
    switch (call.method) {
      case "setReady":
        this.ready = call.arguments['ready'] ?? this.ready;
        _callbackAfterStatusChange(this.ready);
        break;
      case "updateConfigurationHandler":
        _handlerUpdateConfiguration?.call();
        break;
      default:
        break;
    }
  }

  void _handleException(Object error) {
    if (error is PlatformException) {
      PlatformException e = error;
      var message = e.message ?? "";
      message = "Kameleoon SDK: " + message;
      switch (e.code) {
        case "sdk_not_ready":
          throw SDKNotReady(message);
        case "experiment_not_found":
          throw ExperimentNotFound(message);
        case "feature_not_found":
          throw FeatureNotFound(message);
        case "not_targeted":
          throw NotTargeted(message);
        case "not_allocated":
          throw NotAllocated(message);
        case "visitor_code_not_valid":
          throw VisitorCodeNotValid(message);
        case "sitecode_disabled":
          throw SiteCodeDisabled(message);
        case "credentials_not_found":
          throw CredentialsNotFound(message);
        case "variation_not_found":
          throw VariationNotFound(message);
        case "variable_not_found":
          throw VariableNotFound(message: message);
        default:
          throw Exception(e);
      }
    }
  }
}
