import 'kameleoon_data.dart';
import 'kameleoon_client_factory.dart';

/// Abstract class [KameleoonClient], need use [KameleoonClientFactory] to create instance
abstract class KameleoonClient {
  /// For mobile SDKs, the initialization of the Kameleoon Client is not immediate, as it needs to perform a server
  /// call to retrieve the current configuration for all active experiments. It is recommended to check if the SDK is
  /// ready by calling this method before triggering an experiment. Alternatively, you could setup exception catching
  /// around the triggerExperiment() method to look for the KameleoonException.SDKNotReady exception, or you could use
  /// the runWhenReady() method with a callback as detailed in the next paragraph.
  bool? isReady();
  /// For mobile SDKs, the initialization of the Kameleoon Client is not immediate, as it needs to perform a
  /// server call to retrieve the current configuration for all active experiments. The `runWhenReady()` method of
  /// the KameleoonClient class allows to pass a callback that will be executed as soon as the SDK is ready for use.
  /// It also allows the use of a timeout.
  /// The first callback will be executed once the Kameleoon client is ready, and should contain code
  /// triggering an experiment and implementing variations. The second callback will be executed if
  /// the specified timeout happens before the client is initialized. Usually this case should
  /// implement the "reference" variation, as the user will be "out of the experiment" if a timeout takes place.
  void runWhenReady(Function readyCallback, Function failCallback, [Duration? timeout]);

  /// To track conversion, use the `trackConversion()` method. This method requires visitorCode and goalId
  /// to track conversion on this particular goal. In addition, this method also accepts revenue as
  /// a third optional argument to track revenue. The visitorCode should be identical to the one that
  /// was used when triggering the experiment. The `trackConversion()` method doesn't return any value.
  /// This method is non-blocking as the server call is made asynchronously.
  Future<void> trackConversion(String visitorCode, int goalID, [double? revenue]);

  /// To associate various data with the current user, we can use the `addData()` method.
  /// This method requires the `visitorCode` as a first parameter, and then accept a list of additional parameters.
  /// Those additional parameters represent the various Data Types allowed in Kameleoon.
  /// Note that the `addData()` method doesn't return any value and doesn't interact with the Kameleoon
  /// back-end servers by itself. Instead, every data declared is saved for further sending via the `flush()`
  /// method. This allows a minimal number of server calls to be made, as data are usually regrouped in a single
  /// server call triggered by the execution of `flush()`.
  Future<void> addData(String visitorCode, List<Data> data);

  /// Data associated with the current user via `addData()` method is not sent immediately to the server.
  /// It is stored and accumulated until it is sent automatically by the `triggerExperiment()`, `trackConversion()`
  /// `getFeatureVariationKey()`, `getFeatureVariable` or `isFeatureActive` methods, or manually
  /// by the `flush()` method. This allows the developer to exactly control when the data is flushed to our servers.
  /// For instance, if you call the `addData()` method a dozen times, it would be a waste of resources to send data
  /// to the server after each `addData()` invocation. Just call `flush()` once at the end.
  /// The `flush()` method doesn't return any value. This method is non-blocking as the server call
  /// is made asynchronously.
  Future<void> flush([String? visitorCode]);

  /// Running an A/B experiment on your Flutter application means bucketing your users into several
  /// groups (one per variation). The SDK takes care of this bucketing (and the associated reporting) automatically.
  /// Triggering an experiment by calling the `triggerExperiment()` method will register a random variation
  /// for a given visitorCode. If this visitorCode is already associated with a variation (most likely the user
  /// had already been exposed to the experiment previously), then it will return the
  /// previous variation associated with a given experiment.
  Future<int> triggerExperiment(String visitorCode, int experimentID);

  /// Checks if feature is active for a given visitor code
  ///
  /// This method takes a `visitorCode` and `featureKey`
  /// as mandatory arguments to check if the specified feature will be active for a given user.
  /// If such a user has never been associated with this feature flag,
  /// the SDK returns a boolean value randomly (true if the user should have this feature or false if not).
  /// If a user with a given visitorCode is already registered with this feature flag,
  /// it will detect the previous featureFlag value.
  ///
  /// - Parameters:
  ///   - `visitorCode`: Unique identifier of the user. This field is mandatory.
  ///   - `featureKey`: Key of the feature you want to expose to a user. This field is mandatory.
  /// - Throws:
  ///     - `SDKNotReady` if the sdk did not finished its initialization
  ///     - `FeatureFlagNotFound` if the feature was not found in the configuration file
  ///     - `VisitorCodeNotValid` if the visitor code is empty or longer than 255 chars
  /// - Returns:`true` if feature flag is active for a given user, otherwise `false`.
  Future<bool> isFeatureActive(String visitorCode, String featureKey);

  /// Returns a variation key for a given visitor code
  ///
  /// This method takes a `visitorCode` and `featureKey` as mandatory arguments and
  /// returns a variation assigned for a given visitor.
  /// If such a user has never been associated with any feature flag rules, the SDK returns a variation randomly.
  /// If a user with a given `visitorCode` is already registered with this feature flag,
  /// it will detect the previous featureFlag value.
  /// You have to make sure that proper error handling is set up in your code as shown in the example
  /// to the right to catch potential exceptions.
  ///
  /// - Parameters:
  ///   - `visitorCode`: Unique identifier of the user. This field is mandatory.
  ///   - `featureKey`: Key of the feature you want to expose to a user. This field is mandatory.
  /// - Throws: `SDKNotReady` if the sdk did not finished its initialization
  /// - Throws: `FeatureFlagNotFound` if the feature was not found in the configuration file
  /// - Throws: `VisitorCodeNotValid` if the visitor code is empty or longer than 255 chars
  /// - Returns: Variation key assigned for a given visitor
  Future<String> getFeatureVariationKey(String visitorCode, String featureKey);

  /// Retrieves a feature variable value from assigned for a visitor variation
  /// A feature variable can be changed easily via our web application.
  ///
  /// This method takes a `visitorCode`, `featureKey`, `variableKey` as mandatory arguments and
  /// returns a feature variable value for assigned for a given visitor variation.
  /// If such a user has never been associated with any feature flag rules, the SDK
  /// selects a variation randomly and return feature variable value for this variation.
  /// If a user with a given visitorCode is already registered with this feature flag,
  /// it will detect the previous assgined variation for a given visitor.
  /// You have to make sure that proper error handling is set up in your code as shown in the example
  /// to the right to catch potential exceptions.
  ///
  /// - Parameters:
  ///   - visitorCode: Unique identifier of the user. This field is mandatory.
  ///   - featureKey: Key of the feature you want to expose to a user. This field is mandatory.
  /// - Throws: `sdkNotReady` if the sdk did not finished its initialization
  /// - Throws: `featureFlagNotFound` if the feature was not found in the configuration file
  /// - Throws: `variableNotFound` if the variable key was not found in the assigned variation for a given visitor
  /// - Throws: `visitorCodeNotValid` if the visitor code is empty or longer than 255 chars
  /// - Returns: Feature variable value for assigned for a given visitor variation.
  Future<dynamic> getFeatureVariable(String visitorCode, String featureKey, String variableKey);

  /// To retrieve JSON data associated with a variation, call the getVariationAssociatedData() method of our SDK.
  ///
  /// The JSON data usually represents some metadata of the variation,
  /// and can be configured on our web application interface or via our Automation API.
  /// This method takes the variationID as a parameter and will return the data as a [String: Any] instance.
  /// It will throw an exception (`VariationConfigurationNotFound`)
  /// if the variation ID is wrong or corresponds to an experiment that is not yet online.
  ///
  /// - Parameter(s):
  ///     - `variationId` ID of the variation you want to obtain associated data for. This field is mandatory.
  /// - Throws: `VariationNotFound` if the variation was not found in the configuration file
  /// - Returns: Data associated with this `variationId`.
  Future<dynamic> getVariationAssociatedData(int variationId);
  @Deprecated('Use [getVariationAssociatedData] instead')
  Future<dynamic> obtainVariationAssociatedData(int variationId);

  /// Retrieves all feature variables for specific variation key.
  /// A feature variable can be changed easily via our web application.
  ///
  /// This method takes two input parameters: `featureKey` and `variationKey`
  /// It will return the data as a `Future<Map>` instance.
  /// Usually it should be casted to the expected type (the one defined on the web interface).
  /// It will throw an exception (`FeatureFlagNotFound`) if the requested feature has not been found
  /// in the internal configuration of the SDK.
  ///
  /// - Parameters:
  ///   - featureKey: Key of the feature you want to obtain to a user. This field is mandatory.
  ///   - variationKey: Key of the variation you want to obtain to a user. This field is mandatory.
  /// - Throws: `FeatureNotFound` Exception indicating that the requested feature Key has not been found in the
  ///                             internal configuration of the SDK. This is usually normal and means that the feature
  ///                             flag has not yet been activated on Kameleoon's side.
  /// - Throws: `VariationNotFound` Variation key isn't found for current feature flag.
  /// - Returns:  Data associated with this specific variation of this feature flag.
  ///             This can be a int, double, String, bool or Map (depending on the type defined on the web interface).
  Future<Map<String, dynamic>> getFeatureAllVariables(String featureKey, String variationKey);

  /// To retrieve a feature variable, call the `obtainFeatureVariable()` method of our SDK. A feature variable can
  /// be changed easily via our web application.
  /// This method takes two input parameters: `featureKey` and `variableKey`. It will return the data
  /// as an `Object` instance. Usually it should be casted to the expected type (the one defined on the web interface).
  /// It will throw an exception `KameleoonException.FeatureNotFound` if the requested feature has not been found
  /// in the internal configuration of the SDK.
  @Deprecated('Use [getFeatureVariable] instead')
  Future<dynamic> obtainFeatureVariable(var featureKeyOrID, String variableKey);

  /// The `getRemoteData()` method allows you to retrieve data (according to a key passed as argument) for
  /// specified siteCode (specified in `KameleoonClientFactory.create()`) stored on a remote Kameleoon server.
  /// Usually data will be stored on our remote servers via the use of our Data API. This method, along with
  /// the availability of our highly scalable servers for this purpose, provides a convenient way to quickly store
  /// massive amounts of data that can be later retrieved for each of your visitors / users.
  /// Note that since a server call is required, this mechanism is asynchronous.
  Future<dynamic> getRemoteData(String key);
  @Deprecated('Use [getRemoteData] instead')
  Future<dynamic> retrieveDataFromRemoteSource(String key);

  Future<String> getVisitorCode({String? domain, String? defaultVisitorCode});
  @Deprecated('Use [getVisitorCode] instead')
  Future<String?> obtainVisitorCode([String? defaultVisitorCode]);

  /// If real-time configuration updating is enabled, you can call our SDK's `updateConfigurationHandler()` method
  /// to set the update event handler.
  /// This method takes one input parameter: `handler`
  /// The handler is a closure that will be called when the configuration is updated using a real-time configuration event.
  /// - Parameters:
  ///   - handler: handler you want to call on configuration update. This field is mandatory.
  void updateConfigurationHandler(Function handler);

  /// Retrieves a list of experiment IDs for either all existing experiments (if no visitor code is provided) or
  /// specifically targeted or active experiments that are currently active for a given visitor.
  ///
  /// - Parameters:
  ///   - visitorCode: A unique identifier for the visitor. If provided, the method will return experiment IDs
  /// that are specifically targeted to this visitor. If omitted, the method will return experiment IDs
  /// for all existing experiments. The field is optional.
  ///   - onlyAllocated: If set to true, the resulting list will only include experiments where the visitor
  /// has been allocated traffic. Set this parameter to false to fetch all targeted experiments,
  /// regardless of traffic allocation. By default, the value has `true` value.
  /// - Throws: `VisitorCodeNotValid` if the visitor code is empty or longer than 255 chars
  /// - Returns: A list of experiment IDs, either encompassing all experiments or targeted experiments or active
  /// for the specified visitorCode.
  Future<List<int>> getExperimentList([String visitorCode, bool onlyAllocated]);

  /// Returns a list of feature flag keys currently available for the sSDK
  ///
  /// - Returns: List of all feature flag keys
  Future<List<String>> getFeatureList();

  /// Returns a list of function flag keys that are active for a given visitor
  ///
  /// - Parameters:
  ///   - visitorCode: A unique identifier for the visitor.
  /// - Throws: `VisitorCodeNotValid` if the visitor code is empty or longer than 255 chars
  /// - Returns: List of feature keys of active feature flags for current visitorCode.
  Future<List<String>> getFeatureListActive(String visitorCode);
}
